@@ -1,9 +1,12 @@
+from __future__ import print_function
+
 import os
 import sys
 if sys.version_info >= (3,):
     import http.client
 else:
     import httplib
+from requests.exceptions import HTTPError
 import mimetypes
 import zipfile
 import time
@@ -55,6 +58,7 @@ class Task:
     priority = 0
     label = None
     cluster = None
+    sync = False
 
     scheduled = False
     repeating = False
@@ -238,8 +242,8 @@ class CodePackage:
 
 class IronWorker:
     NAME = "iron_worker_python"
-    VERSION = "1.3.5"
-    
+    VERSION = "1.3.8"
+
     isLoaded = False
     arguments = {'task_id': None, 'dir': None, 'payload': None, 'config': None}
 
@@ -249,7 +253,7 @@ class IronWorker:
         Keyword arguments are passed directly to iron_core_python; consult its
         documentation for a full list and possible values."""
         self.client = iron_core.IronClient(name=IronWorker.NAME,
-                version=IronWorker.VERSION, product="iron_worker", **kwargs)    
+                version=IronWorker.VERSION, product="iron_worker", **kwargs)
 
     #############################################################
     ####################### CODE PACKAGES #######################
@@ -369,18 +373,63 @@ class IronWorker:
     #############################################################
     ########################## TASKS ############################
     #############################################################
-    def tasks(self, scheduled=False, page=None, per_page=30):
+    def tasks(self, scheduled=False, page=None, per_page=30, **kwargs):
+        query_params = []
+
+        if page:
+            query_params.append("page=%s" % page)
+
+        if per_page > 100:
+            raise Exception("The limit for per_page is 100")
+
+        elif per_page:
+            query_params.append('per_page=%s' % per_page)
+
         if not scheduled:
-            resp = self.client.get("tasks")
+            if 'queued' in kwargs:
+                query_params.append('queued=1')
+
+            if 'running' in kwargs:
+                query_params.append('running=1')
+
+            if 'complete' in kwargs:
+                query_params.append('complete=1')
+
+            if 'error' in kwargs:
+                query_params.append('error=1')
+
+            if 'cancelled' in kwargs:
+                query_params.append('cancelled=1')
+
+            if 'killed' in kwargs:
+                query_params.append('killed=1')
+
+            if 'timeout' in kwargs:
+                query_params.append('timeout=1')
+
+            if 'from_time' in kwargs:
+                if not (isinstance(kwargs['from_time'], int)):
+                    raise Exception('The time should be epoch integer seconds')
+
+                query_params.append('from_time=%s' % kwargs['from_time'])
+
+            if 'to_time' in kwargs:
+                if not (isinstance(kwargs['to_time'], int)):
+                    raise Exception('The time should be epoch integer seconds')
+
+                query_params.append('to_time=%s' % kwargs['to_time'])
+
+            if 'code_name' in kwargs:
+                query_params.append('code_name=%s' % kwargs['code_name'])
+
+            query_params = "&".join(query_params)
+            request_string = "tasks?" + query_params
+            resp = self.client.get(request_string)
             raw_tasks = resp["body"]
             raw_tasks = raw_tasks["tasks"]
         else:
-            if per_page > 100:
-                raise Exception("The limit for per_page is 100")
-            query_params = "?per_page=" + str(per_page)
-            if page is not None:
-                query_params += "&page=" + page
-            request_string = "schedules" + query_params
+            query_params = "&".join(query_params)
+            request_string = "schedules?" + query_params
             resp = self.client.get(request_string)
             raw_tasks = resp["body"]
             raw_tasks = raw_tasks["schedules"]
@@ -419,7 +468,8 @@ class IronWorker:
                     "priority": task.priority,
                     "delay": task.delay,
                     "label": task.label,
-                    "cluster": task.cluster
+                    "cluster": task.cluster,
+                    "sync": task.sync
             }
             if not task.scheduled:
                 type_str = "tasks"
@@ -447,7 +497,7 @@ class IronWorker:
             tasks_data.append(task_data)
         data = json.dumps({type_str: tasks_data})
         headers = {"Content-Type": "application/json"}
-        
+
         if retry is not None:
             resp = self.client.post(type_str, body=data, headers=headers, retry=retry)
         else:
@@ -458,7 +508,7 @@ class IronWorker:
                     for task in tasks[type_str]]
         else:
             return Task(tasks[type_str][0],
-                    scheduled=(type_str == "schedules"))
+                        scheduled=(type_str == "schedules"))
 
     def task(self, id, scheduled=False):
         if isinstance(id, Task):
@@ -481,28 +531,38 @@ class IronWorker:
         headers = {"Accept": "text/plain"}
         resp = self.client.get(url, headers=headers)
         return resp["body"]
-	
+
+    def stdout(self, id):
+        if isinstance(id, Task):
+            if id.scheduled:
+                raise ValueError("Cannot retrieve a scheduled task's stdout.")
+            id = id.id
+        url = "tasks/%s/outlog" % id
+        headers = {"Accept": "text/plain"}
+        resp = self.client.get(url, headers=headers)
+        return resp["body"]
+
     def setProgress(self, id, percent, msg=''):
         if isinstance(id, Task):
             id = id.id
         url = "tasks/%s/progress" % id
-	body = {}
-	body['percent'] = percent
-	body['msg'] = msg
+        body = {}
+        body['percent'] = percent
+        body['msg'] = msg
         body = json.dumps(body)
-	resp = self.client.post(url, body=body,
-                                    headers={"Content-Type":"application/json"})
+        resp = self.client.post(url, body=body,
+                                headers={"Content-Type": "application/json"})
         return resp["body"]
-		
+
     def retry(self, id, delay=1):
         if isinstance(id, Task):
             id = id.id
         url = "tasks/%s/retry" % id
-	body = {}
+        body = {}
         body['delay'] = delay
         body = json.dumps(body)
-	resp = self.client.post(url, body=body,
-                                    headers={"Content-Type":"application/json"})
+        resp = self.client.post(url, body=body,
+                                headers={"Content-Type":"application/json"})
         return resp["body"]
 
     def cancel(self, id, scheduled=False):
@@ -516,6 +576,34 @@ class IronWorker:
         resp = self.client.post(url)
         return True
 
+    def run(self, **kwargs):
+        kwargs['sync'] = True
+        task = self.queue(**kwargs)
+
+        task = self.wait_for_task(task)
+
+        return self.wait_for_task_stdout(task)
+
+    def wait_for_task(self, task):
+        sleep = 0.5
+        task = self.task(task)
+        tries = task.timeout + 15
+        while task.status in ['queued', 'preparing', 'running']:
+            tries -= 1
+            if tries < 0:
+                raise RuntimeError('Timeout waiting for task execution')
+            time.sleep(sleep)
+            sleep = IronWorker.sleep_between_retries(sleep)
+            task = self.task(task)
+        return task
+
+    def wait_for_task_log(self, task):
+        return IronWorker.wait_for_task_output(task, self.log)
+
+    def wait_for_task_stdout(self, task):
+        return IronWorker.wait_for_task_output(task, self.stdout)
+
+
     #############################################################
     ######################### HELPERS ###########################
     #############################################################
@@ -567,20 +655,20 @@ class IronWorker:
                 IronWorker.arguments['payload_file'] = sys.argv[i + 1]
             if sys.argv[i] == "-config":
                 IronWorker.arguments['config_file'] = sys.argv[i + 1]
-        
+
         if os.getenv('TASK_ID'): IronWorker.arguments['task_id'] = os.getenv('TASK_ID')
         if os.getenv('TASK_DIR'): IronWorker.arguments['dir'] = os.getenv('TASK_DIR')
         if os.getenv('PAYLOAD_FILE'): IronWorker.arguments['payload_file'] = os.getenv('PAYLOAD_FILE')
         if os.getenv('CONFIG_FILE'): IronWorker.arguments['config_file'] = os.getenv('CONFIG_FILE')
 
         if 'payload_file' in IronWorker.arguments and file_exists(IronWorker.arguments['payload_file']):
-	    f = open(IronWorker.arguments['payload_file'], "r")
+            f = open(IronWorker.arguments['payload_file'], "r")
             try:
                 content = f.read()
                 f.close()
                 IronWorker.arguments['payload'] = json.loads(content)
-            except Exception, e:
-                print "Couldn't parse IronWorker payload into json, leaving as string. %s" % e
+            except Exception as e:
+                print("Couldn't parse IronWorker payload into json, leaving as string. %s" % e)
 
         if 'config_file' in IronWorker.arguments and file_exists(IronWorker.arguments['config_file']):
             f = open(IronWorker.arguments['config_file'])
@@ -588,8 +676,8 @@ class IronWorker:
                 content = f.read()
                 f.close()
                 IronWorker.arguments['config'] = json.loads(content)
-            except Exception, e:
-                print "Couldn't parse IronWorker config into json. %s" % e
+            except Exception as e:
+                print("Couldn't parse IronWorker config into json. %s" % e)
 
         IronWorker.isLoaded = True
 
@@ -597,7 +685,7 @@ class IronWorker:
     def payload():
         IronWorker.load_args()
         return IronWorker.arguments['payload']
-    
+
     @staticmethod
     def config():
         IronWorker.load_args()
@@ -621,3 +709,26 @@ class IronWorker:
     @staticmethod
     def loaded():
         return IronWorker.isLoaded
+
+    @staticmethod
+    def sleep_between_retries(previous_duration):
+        if previous_duration >= 60:
+            return previous_duration
+        return previous_duration * 2
+
+    @staticmethod
+    def wait_for_task_output(task, out_func):
+        sleep = 0.5
+        tries = 60
+        while True:
+            try:
+                return out_func(task)
+            except HTTPError as e:
+                if e.response.status_code == 404:
+                    tries -= 1
+                    if tries < 0:
+                        raise RuntimeError('Timeout waiting for task output')
+                    time.sleep(sleep)
+                    sleep = IronWorker.sleep_between_retries(sleep)
+                else:
+                    raise e
